@@ -32,12 +32,25 @@ from alphafold3_pytorch.alphafold3 import (
     Alphafold3
 )
 
+from alphafold3_pytorch.inputs import (
+    register_input_transform
+)
+
 from alphafold3_pytorch.trainer import (
     Trainer,
     DataLoader,
     AtomInput
 )
 
+from alphafold3_pytorch.configs import (
+    Alphafold3Config,
+    TrainerConfig,
+    ConductorConfig,
+    create_alphafold3_from_yaml,
+    create_trainer_from_yaml,
+    create_trainer_from_conductor_yaml
+)
+
 __all__ = [
     Attention,
     Attend,
@@ -66,6 +79,12 @@ __all__ = [
     ConfidenceHead,
     DistogramHead,
     Alphafold3,
+    Alphafold3Config,
     AtomInput,
-    Trainer
+    Trainer,
+    TrainerConfig,
+    ConductorConfig,
+    create_alphafold3_from_yaml,
+    create_trainer_from_yaml,
+    create_trainer_from_conductor_yaml
 ]
