@@ -165,6 +165,7 @@ class Attention(Module):
         query_bias = True,
         flash = True,
         window_size = None,
+        num_memory_kv: int = 0,
         efficient_attn_config: Config = Config(True, True, True)
     ):
         super().__init__()
@@ -178,6 +179,7 @@ class Attention(Module):
         e - dimension (pairwise rep)
         i - source sequence
         j - context sequence
+        m - memory key / value seq
         """
 
         dim_inner = dim_head * heads
@@ -196,6 +198,12 @@ class Attention(Module):
         self.to_kv = nn.Linear(dim, dim_inner * 2, bias = False)
         self.to_out = nn.Linear(dim_inner, dim, bias = False)
 
+        self.memory_kv = None
+
+        if num_memory_kv > 0:
+            self.memory_kv = nn.Parameter(torch.zeros(2, heads, num_memory_kv, dim_head))
+            nn.init.normal_(self.memory_kv, std = 0.02)
+
         # gating of value
         # allows attention to attend to nothing
 
@@ -230,7 +238,8 @@ class Attention(Module):
         out = self.attend(
             q, k, v,
             attn_bias = attn_bias,
-            mask = mask
+            mask = mask,
+            memory_kv = self.memory_kv
         )
 
         # merge heads
@@ -315,7 +324,8 @@ class Attend(Module):
         k: Float['b h n d'],
         v: Float['b h n d'],
         mask: Bool['b n'] | None = None,
-        attn_bias: Float['... n n'] | Float['... nw w (w*2)'] | None = None
+        attn_bias: Float['... n n'] | Float['... nw w (w*2)'] | None = None,
+        memory_kv: Float['2 h m d'] | None = None
     ) -> Float['b h n d']:
         """
         simple local attention with a radius of 1 window size
@@ -363,6 +373,24 @@ class Attend(Module):
 
         q = q * scale
 
+        # append memory key / values for local attention windows
+
+        if exists(memory_kv):
+            batch, seq, num_mem_kv = k.shape[0], k.shape[2], memory_kv.shape[-2]
+
+            mk, mv = memory_kv
+            mk, mv = tuple(repeat(t, 'h m d -> b h n m d', b = batch, n = seq) for t in (mk, mv))
+            k = torch.cat((mk, k), dim = -2)
+            v = torch.cat((mv, v), dim = -2)
+
+            if exists(attn_bias):
+                attn_bias = pad_at_dim(attn_bias, (num_mem_kv, 0), value = 0.)
+
+            if exists(mask):
+                mask = pad_at_dim(mask, (num_mem_kv, 0), value = True)
+
+        # similarity
+
         sim = einsum(q, k, "... i d, ... j d -> ... i j")
 
         if exists(attn_bias):
@@ -399,6 +427,7 @@ class Attend(Module):
         v: Float['b h j d'],
         mask: Bool['b j'] | None = None,
         attn_bias: Float['... i j'] | Float['... nw w (w*2)'] | None = None,
+        memory_kv: Float['2 h m d'] | None = None
     ) -> Float['b h i d']:
 
         is_windowed_attn_bias = None
@@ -410,10 +439,26 @@ class Attend(Module):
         # todo (handle attn bias efficiently)
 
         if self.is_local_attn:
-            return self.local_attn(q, k, v, mask = mask, attn_bias = attn_bias)
+            return self.local_attn(q, k, v, mask = mask, attn_bias = attn_bias, memory_kv = memory_kv)
 
         assert not exists(is_windowed_attn_bias) or not is_windowed_attn_bias
 
+        # append memory key / values
+
+        if exists(memory_kv):
+            batch, num_mem_kv = q.shape[0], memory_kv.shape[-2]
+
+            mk, mv = memory_kv
+            mk, mv = tuple(repeat(t, 'h m d -> b h m d', b = batch) for t in (mk, mv))
+            k = torch.cat((mk, k), dim = -2)
+            v = torch.cat((mv, v), dim = -2)
+
+            if exists(attn_bias):
+                attn_bias = pad_at_dim(attn_bias, (num_mem_kv, 0), value = 0.)
+
+            if exists(mask):
+                mask = pad_at_dim(mask, (num_mem_kv, 0), value = True)
+
         # forward to using flash attention if applicable
 
         can_use_flash = self.flash and not exists(attn_bias), 'flash attention does not support attention bias with gradients'
