@@ -3,6 +3,7 @@ os.environ['TYPECHECK'] = 'True'
 
 import torch
 import pytest
+from pathlib import Path
 
 from alphafold3_pytorch import (
     SmoothLDDTLoss,
@@ -21,7 +22,12 @@ from alphafold3_pytorch import (
     InputFeatureEmbedder,
     ConfidenceHead,
     DistogramHead,
-    Alphafold3,
+    Alphafold3
+)
+
+from alphafold3_pytorch.configs import (
+    Alphafold3Config,
+    create_alphafold3_from_yaml
 )
 
 from alphafold3_pytorch.alphafold3 import (
@@ -262,6 +268,9 @@ def test_diffusion_module():
         atom_encoder_depth = 1,
         atom_decoder_depth = 1,
         token_transformer_depth = 1,
+        atom_encoder_kwargs = dict(
+            attn_num_memory_kv = 2
+        ),
         token_transformer_kwargs = dict(
             num_register_tokens = 2
         )
@@ -355,7 +364,7 @@ def test_confidence_head():
 
     confidence_head = ConfidenceHead(
         dim_single_inputs = 77,
-        atompair_dist_bins = torch.linspace(3, 20, 37),
+        atompair_dist_bins = torch.linspace(3, 20, 37).tolist(),
         dim_single = 384,
         dim_pairwise = 128,
     )
@@ -398,8 +407,10 @@ def test_distogram_head():
     logits = distogram_head(pairwise_repr)
 
 @pytest.mark.parametrize('window_atompair_inputs', (True, False))
+@pytest.mark.parametrize('stochastic_frame_average', (True, False))
 def test_alphafold3(
-    window_atompair_inputs: bool
+    window_atompair_inputs: bool,
+    stochastic_frame_average: bool
 ):
     seq_len = 16
     atoms_per_window = 27
@@ -454,6 +465,7 @@ def test_alphafold3(
             token_transformer_depth = 1,
             atom_decoder_depth = 1,
         ),
+        stochastic_frame_average = stochastic_frame_average
     )
 
     loss, breakdown = alphafold3(
@@ -550,3 +562,153 @@ def test_alphafold3_without_msa_and_templates():
     )
 
     loss.backward()
+
+def test_alphafold3_force_return_loss():
+    seq_len = 16
+    molecule_atom_lens = torch.randint(1, 3, (2, seq_len))
+    atom_seq_len = molecule_atom_lens.sum(dim = -1).amax()
+
+    atom_inputs = torch.randn(2, atom_seq_len, 77)
+    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)
+    additional_molecule_feats = torch.randn(2, seq_len, 10)
+
+    atom_pos = torch.randn(2, atom_seq_len, 3)
+    molecule_atom_indices = molecule_atom_lens - 1
+
+    distance_labels = torch.randint(0, 38, (2, seq_len, seq_len))
+    pae_labels = torch.randint(0, 64, (2, seq_len, seq_len))
+    pde_labels = torch.randint(0, 64, (2, seq_len, seq_len))
+    plddt_labels = torch.randint(0, 50, (2, seq_len))
+    resolved_labels = torch.randint(0, 2, (2, seq_len))
+
+    alphafold3 = Alphafold3(
+        dim_atom_inputs = 77,
+        dim_template_feats = 44,
+        num_dist_bins = 38,
+        confidence_head_kwargs = dict(
+            pairformer_depth = 1
+        ),
+        template_embedder_kwargs = dict(
+            pairformer_stack_depth = 1
+        ),
+        msa_module_kwargs = dict(
+            depth = 1
+        ),
+        pairformer_stack = dict(
+            depth = 2
+        ),
+        diffusion_module_kwargs = dict(
+            atom_encoder_depth = 1,
+            token_transformer_depth = 1,
+            atom_decoder_depth = 1,
+        ),
+    )
+
+    sampled_atom_pos = alphafold3(
+        num_recycling_steps = 2,
+        atom_inputs = atom_inputs,
+        molecule_atom_lens = molecule_atom_lens,
+        atompair_inputs = atompair_inputs,
+        additional_molecule_feats = additional_molecule_feats,
+        atom_pos = atom_pos,
+        molecule_atom_indices = molecule_atom_indices,
+        distance_labels = distance_labels,
+        pae_labels = pae_labels,
+        pde_labels = pde_labels,
+        plddt_labels = plddt_labels,
+        resolved_labels = resolved_labels,
+        return_loss_breakdown = True,
+        return_loss = False # force sampling even if labels are given
+    )
+
+    assert sampled_atom_pos.ndim == 3
+
+    loss, _ = alphafold3(
+        num_recycling_steps = 2,
+        atom_inputs = atom_inputs,
+        molecule_atom_lens = molecule_atom_lens,
+        atompair_inputs = atompair_inputs,
+        additional_molecule_feats = additional_molecule_feats,
+        return_loss_breakdown = True,
+        return_loss = True # force returning loss even if no labels given
+    )
+
+    assert loss == 0.
+
+def test_alphafold3_with_atom_and_bond_embeddings():
+    alphafold3 = Alphafold3(
+        num_atom_embeds = 7,
+        num_atompair_embeds = 3,
+        dim_atom_inputs = 77,
+        dim_template_feats = 44
+    )
+
+    # mock inputs
+
+    seq_len = 16
+
+    molecule_atom_lens = torch.randint(1, 3, (2, seq_len))
+    atom_seq_len = molecule_atom_lens.sum(dim = -1).amax()
+
+    atom_ids = torch.randint(0, 7, (2, atom_seq_len))
+    atompair_ids = torch.randint(0, 3, (2, atom_seq_len, atom_seq_len))
+
+    atom_inputs = torch.randn(2, atom_seq_len, 77)
+    atompair_inputs = torch.randn(2, atom_seq_len, atom_seq_len, 5)
+
+    additional_molecule_feats = torch.randn(2, seq_len, 10)
+
+    template_feats = torch.randn(2, 2, seq_len, seq_len, 44)
+    template_mask = torch.ones((2, 2)).bool()
+
+    msa = torch.randn(2, 7, seq_len, 64)
+    msa_mask = torch.ones((2, 7)).bool()
+
+    # required for training, but omitted on inference
+
+    atom_pos = torch.randn(2, atom_seq_len, 3)
+    molecule_atom_indices = molecule_atom_lens - 1 # last atom, as an example
+
+    distance_labels = torch.randint(0, 37, (2, seq_len, seq_len))
+    pae_labels = torch.randint(0, 64, (2, seq_len, seq_len))
+    pde_labels = torch.randint(0, 64, (2, seq_len, seq_len))
+    plddt_labels = torch.randint(0, 50, (2, seq_len))
+    resolved_labels = torch.randint(0, 2, (2, seq_len))
+
+    # train
+
+    loss = alphafold3(
+        num_recycling_steps = 2,
+        atom_ids = atom_ids,
+        atompair_ids = atompair_ids,
+        atom_inputs = atom_inputs,
+        atompair_inputs = atompair_inputs,
+        molecule_atom_lens = molecule_atom_lens,
+        additional_molecule_feats = additional_molecule_feats,
+        msa = msa,
+        msa_mask = msa_mask,
+        templates = template_feats,
+        template_mask = template_mask,
+        atom_pos = atom_pos,
+        molecule_atom_indices = molecule_atom_indices,
+        distance_labels = distance_labels,
+        pae_labels = pae_labels,
+        pde_labels = pde_labels,
+        plddt_labels = plddt_labels,
+        resolved_labels = resolved_labels
+    )
+
+    assert loss.numel() == 1
+
+# test creation from config
+
+def test_alphafold3_config():
+    curr_dir = Path(__file__).parents[0]
+    af3_yaml = curr_dir / 'configs/alphafold3.yaml'
+    trainer_yml = curr_dir / 'configs/trainer.yaml'
+
+    alphafold3 = create_alphafold3_from_yaml(af3_yaml)
+    assert isinstance(alphafold3, Alphafold3)
+
+    alphafold3_from_trainer_yml = create_alphafold3_from_yaml(trainer_yml, 'model')
+    assert isinstance(alphafold3_from_trainer_yml, Alphafold3)
